#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import os
import sys
import requests
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.projections.polar import PolarAxes
import networkx as nx
from scipy.spatial import distance_matrix
from astropy.coordinates import SkyCoord, EarthLocation, AltAz  # type: ignore
from astropy.time import Time  # type: ignore
from astropy import units as u  # type: ignore
from typing import cast
import datetime


class CelestialTSP:
    def __init__(self, args: argparse.Namespace):
        self.args = args
        self.location = EarthLocation(lat=args.lat, lon=args.lon, height=args.height)
        self.location_str = f"Lat: {args.lat}, Lon: {args.lon}, Height: {args.height}m"
        # If a “default date/time” was provided then no timezone conversion is required.
        if args.default_datetime:
            self.observation_time = Time(f"{args.date} {args.time}")
        else:
            self.observation_time = (
                Time(f"{args.date} {args.time}") - int(args.tz) * u.hour
            )
        self.df: pd.DataFrame = pd.DataFrame()

    def run(self):
        # 1. Read celestial names/coordinates from file.
        self.df = self.read_celestial_names(self.args.input_file_path)

        # 2. Check if first_body name is provided and valid.
        if self.args.first_body:
            # Instead of int(self.df.index[...][0]), we explicitly get the index.
            matching_indices = self.df.index[self.df["Name"] == self.args.first_body]
            if matching_indices.empty:
                print(f"Error: {self.args.first_body} is not in the input file.")
                sys.exit(1)
            first_index = int(matching_indices[0])  # type: ignore
        else:
            first_index = self.find_first_body()

        print(f"Location: {self.location_str}")
        print(
            f"Observation Date/Time: {self.args.date} {self.args.time} {self.args.tz}"
        )

        # 3. Build a graph for all celestial bodies:
        # We use their (Altitude, Azimuth) values.
        coordinates = self.df[["Altitude", "Azimuth"]].to_numpy()
        dmatrix = distance_matrix(coordinates, coordinates)
        if first_index != -1:
            graph = self.make_graph(coordinates, dmatrix)
            tsp_path = nx.approximation.greedy_tsp(graph, source=first_index)
            df_ordered = self.df.iloc[tsp_path].reset_index(drop=True)
            self.show_results(df_ordered)
            self.save_spherical_image(
                df_ordered, self.location_str, self.observation_time, self.args.output
            )
        else:
            print(
                "Could not find the celestial body closest to the west (270° azimuth)."
            )

    def read_celestial_names(self, file_path: str) -> pd.DataFrame:
        """
        Reads a file with celestial object names and retrieves their RA/Dec using astropy.
        Exits with error if file cannot be found or no valid data is returned.
        """
        if not os.path.exists(file_path):
            print(f"Input file {file_path} does not exist.")
            sys.exit(1)

        records = []
        with open(file_path, "r") as f:
            for line in f:
                name = line.strip()
                try:
                    coord = SkyCoord.from_name(name)
                    records.append(
                        {"Name": name, "RA": coord.ra.deg, "Dec": coord.dec.deg}
                    )
                except Exception as e:
                    print(f"Error looking up {name}: {e}")
                    continue
        df = pd.DataFrame(records)
        if df.empty:
            print("Input file is empty or contains no valid celestial names.")
            sys.exit(1)
        return df

    def find_first_body(self) -> int:
        """
        Identify the celestial body that will set first (i.e. has the shortest time until setting)
        when observed from self.location at self.observation_time.
        The method also adds several columns to self.df: Altitude, Azimuth, TimeToSet, Observable.
        Returns the row index of this first body.
        """
        altaz_frame = AltAz(obstime=self.observation_time, location=self.location)

        # Pre-create lists (can be replaced with vectorized operations if desired)
        altitudes = []
        azimuths = []
        times_to_set = []
        observables = []
        shortest_time = np.inf
        first_index = -1

        # For each row we can optimize by processing in bulk. However, because the time-to-set
        # calculation uses a simulated time grid, we loop for clarity.
        for i, row in self.df.iterrows():
            sky_coord = SkyCoord(ra=row["RA"], dec=row["Dec"], unit="deg")
            altaz = sky_coord.transform_to(altaz_frame)
            alt = altaz.alt.deg
            az = altaz.az.deg
            altitudes.append(alt)
            azimuths.append(az)
            observables.append(self.is_observable(altaz))

            # If object is currently above horizon compute when it will set
            if alt > 0:
                # Create a time grid (1000 steps within 24h)
                delta_hours = np.linspace(0, 24, 1000) * u.hour
                future_times = self.observation_time + delta_hours
                future_altaz = sky_coord.transform_to(
                    AltAz(obstime=future_times, location=self.location)
                )
                future_alts = future_altaz.alt.deg
                # Find first time when altitude goes non-positive (object sets)
                set_indices = np.where(future_alts <= 0)[0]
                if set_indices.size > 0:
                    t_set = (
                        (future_times[set_indices[0]] - self.observation_time)
                        .to(u.hour)
                        .value
                    )
                    times_to_set.append(t_set)
                    if t_set < shortest_time:
                        shortest_time = t_set
                        first_index = i  # type: ignore
                else:
                    times_to_set.append(np.inf)
            else:
                times_to_set.append(np.inf)

        self.df["Altitude"] = altitudes
        self.df["Azimuth"] = azimuths
        self.df["TimeToSet"] = times_to_set
        self.df["Observable"] = observables

        return first_index

    @staticmethod
    def is_observable(altaz_coord, min_altitude=0) -> bool:
        """Return True if the altitude is above min_altitude (default=0 deg)."""
        return altaz_coord.alt.deg > min_altitude

    @staticmethod
    def make_graph(coordinates: np.ndarray, dist_matrix: np.ndarray) -> nx.Graph:
        """
        Builds a fully-connected NetworkX graph where each node represents a celestial object.
        Node positions (for plotting) are stored in the 'pos' attribute; edge weights are from distance matrix.
        """
        G: nx.Graph = nx.Graph()
        n = len(coordinates)
        for i in range(n):
            G.add_node(i, pos=(coordinates[i][0], coordinates[i][1]))
        # Use upper-triangle of matrix (graph undirected)
        for i in range(n):
            for j in range(i + 1, n):
                G.add_edge(i, j, weight=dist_matrix[i, j])
        return G

    @staticmethod
    def show_results(df: pd.DataFrame) -> None:
        """Prints the optimal order of celestial bodies."""
        print("\nOptimal Order of Celestial Bodies:")
        for i, row in df.iterrows():
            if int(i) + 1 == len(df):  # type: ignore
                continue
            name = row["Name"]
            ra = f"{row['RA']:.2f}"
            dec = f"{row['Dec']:.2f}"
            alt = f"{row['Altitude']:.2f}"
            azimuth = f"{row['Azimuth']:.2f}"
            tset = f"{row['TimeToSet']:.2f}"
            observable = row["Observable"]
            print(
                f"Name: {name:<10} RA: {ra:<7} Dec: {dec:<7} Altitude: {alt:<7} Azimuth: {azimuth:<7} Time to set: {tset:<7} Observable: {observable}"
            )

    @staticmethod
    def save_spherical_image(
        df: pd.DataFrame, location_str: str, observation_time: Time, filename: str
    ) -> None:
        """
        Generate and save a spherical (polar) plot showing celestial object positions
        and the TSP path.
        """
        fig = plt.figure(figsize=(8, 8))
        ax = cast(PolarAxes, fig.add_subplot(111, projection="polar"))

        az_radians = np.deg2rad(df["Azimuth"])
        alt_radians = np.deg2rad(90 - df["Altitude"])

        ax.scatter(az_radians, alt_radians, c="blue", label="Celestial Bodies")
        for i, row in df.iterrows():
            ax.annotate(
                row["Name"], (az_radians[i], alt_radians[i]), fontsize=8, ha="right"
            )

        for i in range(len(df) - 1):
            start_az, start_alt = az_radians[i], alt_radians[i]
            end_az, end_alt = az_radians[i + 1], alt_radians[i + 1]
            ax.plot([start_az, end_az], [start_alt, end_alt], "r-")

        if len(df) > 0:
            start_az, start_alt = az_radians.iloc[0], alt_radians.iloc[0]
            ax.annotate(
                "Start",
                xy=(start_az, start_alt),
                xytext=(start_az, start_alt + 0.1),
                fontsize=12,
                color="green",
                ha="center",
            )
            if len(df) > 1:
                second_az, second_alt = az_radians.iloc[1], alt_radians.iloc[1]
                ax.annotate(
                    "",
                    xy=(second_az, second_alt),
                    xytext=(start_az, start_alt),
                    arrowprops=dict(facecolor="red", arrowstyle="->", lw=2.5),
                )

        ax.set_title("Optimal Order of Celestial Bodies (Spherical Projection)", pad=30)
        ax.set_theta_zero_location("N")
        ax.set_theta_direction(-1)

        obs_time_str = (
            observation_time.iso if observation_time is not None else "Unknown"
        )
        fig.text(
            0.5,
            0.01,
            f"Location: {location_str} | Observation Time: {obs_time_str} UTC",
            ha="center",
            fontsize=10,
        )
        now = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        try:
            plt.savefig(f"{filename}_{now}.png")
            print(f"Plot saved as {filename}_{now}.png")
        except Exception as e:
            print(f"Error saving plot: {e}")
        plt.close()

    @staticmethod
    def get_location() -> tuple:
        """
        Obtain the current location by using ip-api.com.
        Returns a tuple (latitude, longitude) or (None, None) if unavailable.
        """
        try:
            response = requests.get("http://ip-api.com/json/")
            data = response.json()
            if data.get("status") == "success":
                return float(data.get("lat")), float(data.get("lon"))
            else:
                print("Error: Unable to get location data")
                return None, None
        except Exception:
            return None, None

    @classmethod
    def build_arg_parser(cls) -> argparse.Namespace:
        """
        Set up and parse command-line arguments.
        """
        latitude, longitude = cls.get_location()
        # Get current time string for default values.
        now = Time.now().iso.split(".")[0]
        date_default, time_default = now.split(" ")
        parser = argparse.ArgumentParser(description="Celestial TSP Planner")
        parser.add_argument(
            "input_file_path",
            type=str,
            help="Input file path with celestial coordinates.",
        )
        parser.add_argument(
            "--lat",
            type=float,
            default=latitude if latitude is not None else 0,
            help="Latitude of observation location.",
        )
        parser.add_argument(
            "--lon",
            type=float,
            default=longitude if longitude is not None else 0,
            help="Longitude of observation location.",
        )
        parser.add_argument(
            "--height",
            type=float,
            default=0,
            help="Height of observation location (in meters).",
        )
        parser.add_argument(
            "--date",
            type=str,
            default=date_default,
            help="Observation date (YYYY-MM-DD).",
        )
        parser.add_argument(
            "--time",
            type=str,
            default=time_default,
            help="Observation time (HH:MM:SS).",
        )
        parser.add_argument(
            "--tz", type=str, default="+9", help="Time zone offset (e.g., +9 for JST)."
        )
        parser.add_argument(
            "--output",
            type=str,
            default="results",
            help="Filename for the output image.",
        )
        parser.add_argument(
            "--first_body",
            type=str,
            default="",
            help="Name of the celestial body to start the TSP from.",
        )
        args, _ = parser.parse_known_args()
        args.default_datetime = args.date == date_default and args.time == time_default
        return args


def main():
    # Parse command line arguments
    args = CelestialTSP.build_arg_parser()
    # Create and run the CelestialTSP planner
    planner = CelestialTSP(args)
    planner.run()


if __name__ == "__main__":
    main()



import nox

nox.options.python = "3.13"
nox.options.default_venv_backend = "uv"


# @nox.session(python=["3.13"], tags=["lint"])
def lint(session):
    session.install("ruff")
    session.run("uv", "run", "ruff", "check")
    session.run("uv", "run", "ruff", "format")


# @nox.session(python=["3.13"], tags=["mypy"])
def mypy(session):
    session.install(".")
    session.install(
        "mypy", "types-requests", "pandas-stubs", "types-networkx", "scipy-stubs"
    )
    session.run("uv", "run", "mypy", "src")


@nox.session(python=["3.13"], tags=["pytest"])
def pytest(session):
    session.install(".")
    session.install("pytest", "pytest-cov")
    test_files = ["test.py"]
    session.run(
        "uv",
        "run",
        "pytest",
        "--maxfail=1",
        "--cov=celestsp",
        "--cov-report=term",
        *test_files,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import datetime
import io
import os
import sys
import tempfile
import unittest
from contextlib import redirect_stdout
from unittest.mock import patch, MagicMock

import networkx as nx
import numpy as np
import pandas as pd
from astropy.coordinates import SkyCoord
from astropy.time import Time
import astropy.units as u

# Insert the src directory (which contains the macocr package) into sys.path.
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "src"))

# Import the module under test
from celestsp.main import CelestialTSP, main


# This dummy object will emulate the result of SkyCoord.transform_to.
class DummyAltAz:
    def __init__(self, alt):
        # alt is assumed to be a scalar (float) altitude in degrees.
        self.alt = type("dummy_alt", (), {"deg": alt})
        self.az = type("dummy_az", (), {"deg": 100.0})


# A dummy version of SkyCoord.transform_to.
def dummy_transform_to(self, frame):
    """
    This dummy inspects self.ra.deg: if it is less than 1 then returns an altitude of 10,
    otherwise returns an altitude of 5.
    For vectorized calls (when frame.obstime is an array) the altitude will linearly decrease
    from the chosen value to below 0 so that the first time a set condition is met can be computed.
    """
    dummy_alt = 10.0 if self.ra.deg < 1.0 else 5.0
    # If the frame has an "obstime" attribute, check if it is array-like.
    if hasattr(frame, "obstime"):
        try:
            # if frame.obstime is iterable (vectorized call)
            n = len(frame.obstime)
        except TypeError:
            n = None
        if n is not None:
            # Create a linearly decreasing altitude array:
            # altitude will drop from dummy_alt at t=0 to -1 at t=end.
            alts = dummy_alt - (dummy_alt + 1) * np.linspace(0, 1, n)
            dummy_alt_obj = type("DummyAltArray", (), {})()
            dummy_alt_obj.deg = alts
            dummy_az_obj = type("DummyAzArray", (), {})()
            dummy_az_obj.deg = np.full(n, 100.0)
            dummy = type("Dummy", (), {})()
            dummy.alt = dummy_alt_obj
            dummy.az = dummy_az_obj
            return dummy
        else:
            # Scalar call: return a constant dummy object.
            return DummyAltAz(dummy_alt)
    else:
        return DummyAltAz(dummy_alt)


# A dummy version of SkyCoord.from_name for testing read_celestial_names.
def dummy_from_name(name):
    # Return a dummy SkyCoord with RA and Dec based on the name string.
    # For example, if name can be converted to a float then use that value;
    # otherwise use fixed numbers.
    try:
        ra_val = float(name)
    except ValueError:
        ra_val = 15.0
    # Use dec = ra/2 for testing.
    dec_val = ra_val / 2.0
    dummy = SkyCoord(ra=ra_val * u.deg, dec=dec_val * u.deg)
    return dummy


# A dummy requests.get to simulate get_location.
def dummy_requests_get_success(url):
    # Simulate a successful response.
    response = MagicMock()
    response.json.return_value = {"status": "success", "lat": 35.0, "lon": -120.0}
    return response


def dummy_requests_get_fail(url):
    # Simulate a failed location response.
    response = MagicMock()
    response.json.return_value = {"status": "fail"}
    return response


# --- The test class --- #
class TestCelestialTSP(unittest.TestCase):
    def setUp(self):
        # Create dummy arguments similar to build_arg_parser but overriding defaults.
        self.temp_input = tempfile.NamedTemporaryFile(
            delete=False, mode="w", encoding="utf-8"
        )
        # Write some dummy celestial names (each name on its own line).
        # We choose names that when passed to dummy_from_name produce specific RA/Dec values.
        self.temp_input.write(
            "0\n1\nNonNumeric"
        )  # first two numeric, third will use default 15.0.
        self.temp_input.close()
        # Build a dummy argparse.Namespace
        self.args = argparse.Namespace(
            input_file_path=self.temp_input.name,
            lat=0.0,
            lon=0.0,
            height=0.0,
            date="2023-01-01",
            time="00:00:00",
            tz="+0",
            output="test_output",
            first_body="",
            default_datetime=False,
        )
        # Create a CelestialTSP instance with dummy args.
        self.planner = CelestialTSP(self.args)

    def tearDown(self):
        if os.path.exists(self.temp_input.name):
            os.unlink(self.temp_input.name)
        # Remove any files created by save_spherical_image if necessary.
        for f in os.listdir("."):
            if f.startswith(self.args.output) and f.endswith(".png"):
                os.unlink(f)

    @patch("celestsp.main.SkyCoord.from_name", side_effect=dummy_from_name)
    def test_read_celestial_names_success(self, mock_from_name):
        # Test that read_celestial_names returns a DataFrame with valid data.
        df = self.planner.read_celestial_names(self.args.input_file_path)
        self.assertFalse(df.empty)
        self.assertIn("Name", df.columns)
        self.assertIn("RA", df.columns)
        self.assertIn("Dec", df.columns)
        # Three lines in our file.
        self.assertEqual(len(df), 3)

    def test_read_celestial_names_file_not_exist(self):
        # Provide a non-existent file path.
        fake_path = "nonexistent_file.txt"
        with (
            self.assertRaises(SystemExit) as cm,
            patch("sys.stdout", new=io.StringIO()) as fake_out,
        ):
            self.planner.read_celestial_names(fake_path)
        self.assertEqual(cm.exception.code, 1)
        self.assertIn("does not exist", fake_out.getvalue())

    @patch("celestsp.main.SkyCoord.from_name", side_effect=dummy_from_name)
    def test_read_celestial_names_empty(self, mock_from_name):
        # Create an empty temporary file.
        with tempfile.NamedTemporaryFile(
            delete=False, mode="w", encoding="utf-8"
        ) as tmp:
            empty_path = tmp.name
        try:
            with (
                self.assertRaises(SystemExit) as cm,
                patch("sys.stdout", new=io.StringIO()) as fake_out,
            ):
                self.planner.read_celestial_names(empty_path)
            self.assertEqual(cm.exception.code, 1)
            self.assertIn(
                "empty or contains no valid celestial names", fake_out.getvalue()
            )
        finally:
            os.unlink(empty_path)

    def test_is_observable(self):
        # Create a dummy altaz object with alt.deg = 5.
        dummy = DummyAltAz(5)
        self.assertTrue(CelestialTSP.is_observable(dummy))
        # Test below horizon.
        dummy2 = DummyAltAz(-1)
        self.assertFalse(CelestialTSP.is_observable(dummy2))
        # Test with min_altitude argument.
        self.assertFalse(CelestialTSP.is_observable(dummy, min_altitude=6))

    def test_make_graph(self):
        # Prepare simple coordinates and distance matrix.
        coords = np.array([[10, 20], [30, 40], [50, 60]])
        dist_mat = np.array([[0, 1, 2], [1, 0, 3], [2, 3, 0]])
        graph = CelestialTSP.make_graph(coords, dist_mat)
        self.assertIsInstance(graph, nx.Graph)
        # Check nodes and edge weights.
        self.assertEqual(len(graph.nodes), 3)
        self.assertEqual(len(graph.edges), 3)  # complete graph of 3 nodes: 3 edges.
        self.assertAlmostEqual(graph[0][1]["weight"], 1)

    def test_show_results(self):
        # Create a dummy DataFrame with required columns.
        df = pd.DataFrame(
            {
                "Name": ["A", "B"],
                "RA": [10.0, 20.0],
                "Dec": [5.0, 15.0],
                "Altitude": [30.0, 20.0],
                "Azimuth": [40.0, 50.0],
                "TimeToSet": [2.0, 3.0],
                "Observable": [True, False],
            }
        )
        # Capture printed output.
        out = io.StringIO()
        with redirect_stdout(out):
            CelestialTSP.show_results(df)
        result = out.getvalue()
        self.assertIn("Optimal Order of Celestial Bodies", result)
        self.assertIn("A", result)
        self.assertIn("B", result)

    def test_save_spherical_image(self):
        # Create a dummy DataFrame with required columns.
        df = pd.DataFrame(
            {
                "Name": ["A", "B", "C"],
                "RA": [10.0, 20.0, 30.0],
                "Dec": [5.0, 15.0, 25.0],
                "Altitude": [80.0, 70.0, 60.0],
                "Azimuth": [40.0, 50.0, 60.0],
                "TimeToSet": [2.0, 3.0, 4.0],
                "Observable": [True, True, False],
            }
        )
        location_str = "Test Location"
        observation_time = Time("2023-01-01 00:00:00")
        # Patch plt.savefig so no file is written.
        with (
            patch("celestsp.main.plt.savefig") as mock_savefig,
            patch("datetime.datetime") as mock_datetime,
        ):
            mock_datetime.now.return_value = datetime.datetime(2023, 1, 1, 0, 0, 0)
            CelestialTSP.save_spherical_image(
                df, location_str, observation_time, "dummy_output"
            )
            # Assert that savefig was called.
            mock_savefig.assert_called()

    def test_get_location_success(self):
        with patch(
            "celestsp.main.requests.get", side_effect=dummy_requests_get_success
        ):
            lat, lon = CelestialTSP.get_location()
            self.assertEqual(lat, 35.0)
            self.assertEqual(lon, -120.0)

    def test_get_location_fail(self):
        with patch("celestsp.main.requests.get", side_effect=dummy_requests_get_fail):
            lat, lon = CelestialTSP.get_location()
            self.assertIsNone(lat)
            self.assertIsNone(lon)

    def test_build_arg_parser(self):
        # To make sure build_arg_parser returns the expected Namespace,
        # we simulate a command-line call by patching sys.argv.
        args_list = [
            self.temp_input.name,
            "--lat",
            "12.34",
            "--lon",
            "56.78",
            "--height",
            "100",
            "--date",
            "2023-12-31",
            "--time",
            "23:59:59",
            "--tz",
            "+3",
            "--output",
            "my_results.png",
            "--first_body",
            "TestBody",
        ]
        with patch.object(sys, "argv", ["prog"] + args_list):
            with patch(
                "celestsp.main.CelestialTSP.get_location", return_value=(1.0, 2.0)
            ):
                parser_args = CelestialTSP.build_arg_parser()
                self.assertEqual(parser_args.input_file_path, self.temp_input.name)
                self.assertEqual(parser_args.lat, 12.34)
                self.assertEqual(parser_args.lon, 56.78)
                self.assertEqual(parser_args.height, 100)
                self.assertEqual(parser_args.date, "2023-12-31")
                self.assertEqual(parser_args.time, "23:59:59")
                self.assertEqual(parser_args.tz, "+3")
                self.assertEqual(parser_args.output, "my_results.png")
                self.assertEqual(parser_args.first_body, "TestBody")

    @patch("celestsp.main.SkyCoord.from_name", side_effect=dummy_from_name)
    @patch("celestsp.main.SkyCoord.transform_to", new=dummy_transform_to)
    @patch("celestsp.main.nx.approximation.greedy_tsp", return_value=[1, 0, 2])
    def test_run_with_first_body_not_specified(self, mock_tsp, mock_from_name):
        # Test the full run() method when first_body is not provided.
        # First, read celestial names.
        self.planner.df = self.planner.read_celestial_names(self.args.input_file_path)
        # We expect find_first_body to use our dummy transform_to.
        # With our dummy, rows with RA < 1 get altitude 10 and others get altitude 5.
        # In our test file, the first entry is "0", so RA = 0 and altitude=10; the second is "1"
        # (RA = 1 gives altitude=5) and third is "NonNumeric" -> RA =15 so altitude=5.
        # When computing t_set in find_first_body the one with lower dummy altitude (5) will set earlier.
        out = io.StringIO()
        with redirect_stdout(out):
            # Because run() eventually calls sys.exit(1) if first_body provided is invalid.
            # In this run we'll not provide first_body so find_first_body returns an index.
            self.planner.run()
        # In our dummy simulation, first_index should be 1 (the second row).
        printed = out.getvalue()
        self.assertIn("Location:", printed)
        self.assertIn("Observation Date/Time:", printed)
        # Because all plotting and file saving functions are exercised, check that the output contains the plot saved message.
        self.assertIn("Plot saved as", printed)

    @patch("celestsp.main.SkyCoord.from_name", side_effect=dummy_from_name)
    def test_run_with_invalid_first_body(self, mock_from_name):
        # Set a first_body that is not in the file.
        self.args.first_body = "NonExistent"
        self.planner = CelestialTSP(self.args)
        # read the file (will read three lines and get dataframe)
        self.planner.df = self.planner.read_celestial_names(self.args.input_file_path)
        out = io.StringIO()
        with redirect_stdout(out), self.assertRaises(SystemExit):
            self.planner.run()
        printed = out.getvalue()
        self.assertIn("is not in the input file.", printed)


# Test the module level main() function.
class TestMainFunction(unittest.TestCase):
    @patch("celestsp.main.CelestialTSP.build_arg_parser")
    @patch("celestsp.main.CelestialTSP.run")
    def test_main(self, mock_run, mock_build_arg_parser):
        # Create a dummy Namespace to be returned by build_arg_parser.
        dummy_args = argparse.Namespace(
            input_file_path="dummy.txt",
            lat=0.0,
            lon=0.0,
            height=0.0,
            date="2023-01-01",
            time="00:00:00",
            tz="+0",
            output="dummy_output",
            first_body="",
            default_datetime=True,
        )
        mock_build_arg_parser.return_value = dummy_args
        # Run main() and make sure run() is called.
        with patch("sys.stdout", new=io.StringIO()):
            main()
            mock_run.assert_called()


if __name__ == "__main__":
    unittest.main()
